#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <openssl/evp.h>

#define SA struct sockaddr

void printbyte(char b) {
  
  char c;

  c = b;
  c = c >> 4;
  c = c & 15;
  printf("%X", c);
  c = b;
  c = c & 15;
  printf("%X:", c);
  
}

int retrieve_key(char* key, const int key_size) {
  
   int ret;
   FILE* file;
   struct stat info;
   
   ret = stat("key", &info);
   if (ret != 0)
     return 1;
   
   file = fopen("key", "r");
   if(!file)
      return 1;
   
   ret = fread(key, 1, key_size, file);
   if(ret < key_size)
     return 1;
   
   fclose(file);
   
   return 0;
  
}

int create_enc_context(EVP_CIPHER_CTX *ctx, int* block_size) {
  
   int key_size; /* dimensione chiave di cifratura */
   char* key; /* chiave di cifratura */
      
   /* inizializzazione del contesto */
   EVP_CIPHER_CTX_init(ctx);
  
   /* setup del contesto per la cifratura */
   EVP_EncryptInit(ctx, EVP_des_ecb(), NULL, NULL);
  
   key_size = EVP_CIPHER_CTX_key_length(ctx);
   key = malloc(key_size);
   
   if (retrieve_key(key, key_size)) {
      printf("Errore recupero chiave\n");
      return 1;
   }
  
   /* impostazione della chiave */
   EVP_CIPHER_CTX_set_key_length(ctx, key_size);
   EVP_EncryptInit(ctx, NULL, (unsigned char*) key, NULL);
    
   /* recupero dimensione del blocco */
   *block_size = EVP_CIPHER_CTX_block_size(ctx);
  
   free(key);
   return 0;
  
}
  
int encrypt(EVP_CIPHER_CTX *ctx, const char* buffer, const int size, char* enc_buffer, int* ct_size) {
  
  int nc; /* numero di byte crittati ad ogni passo */
  int nctot; /* numero totale di byte crittati */
  int i; /* indice */
  int ct_ptr, msg_ptr; /* puntatori alla prima posizione libera del buffer */
  int n;  /* numero di byte da cifrare ad ogni passo */
    
  n = 10;
  nc = 0;
  nctot = 0;
  ct_ptr = 0;
  msg_ptr =0;
      
  /* Cifratura a blocchi di n byte */
  /*for (i = 0; i < size / n; i++) {
    EVP_EncryptUpdate(ctx, &enc_buffer[ct_ptr], &nc, &buffer[msg_ptr], n);
    ct_ptr += nc;
    msg_ptr += n;
    nctot += nc;
  }
   
  if (size % n) {
    EVP_EncryptUpdate(ctx, &enc_buffer[ct_ptr], &nc, &buffer[msg_ptr], size % n);
    ct_ptr += nc;
    msg_ptr += size % n;
    nctot += nc;
  }*/
  
  /* Cifratura in un solo passaggio */
  EVP_EncryptUpdate(ctx, enc_buffer, &nc, buffer, size);  
  ct_ptr += nc;
  nctot += nc;
  
  EVP_EncryptFinal(ctx, &enc_buffer[ct_ptr], &nc);
  nctot += nc;
  
  *ct_size = nctot;
  
  return 0;
  
}

   /* Funzione di invio file dal client al server
   @ input file_name = nome del file da inviare
   @ input sk = socket attraverso la quale si invia il file
   @ return 0 in caso di successo, 1 altrimenti */
   int send_file(char* file_name, int sk) {
	  
	FILE* file;			// puntatore al file da inviare
	int name_size;			// dimensione del nome del file da inviare
	int size; 			// dimensione del file da inviare
	int ret, i; 			
	unsigned char* buffer;		// puntatore al buffer contenente il file
	unsigned char* enc_buffer;	// buffer per il testo cifrato
	int enc_buffer_size;		// dimensione del buffer per il testo cifrato
	int block_size;			// dimensione del blocco
	EVP_CIPHER_CTX *ctx; 		// contesto
	int ct_size;			// dimensione del testo cifrato
	
	/* Calcolo della lunghezza del nome del file*/
	name_size = strlen(file_name);

	/* Apertura del file da inviare */
	file = fopen(file_name,"r");
	if(file == NULL) {
	  printf("\nErrore durante l'apertura del file\n");
	  return 1;
	}
    	
    	/* Ricerca della lunghezza del file */
	fseek(file,0,SEEK_END);
	size = ftell(file);
	
	/* Allocazione della memoria per il file */
	buffer = malloc(size * sizeof (char));
	fseek(file, 0, SEEK_SET);

	/* Lettura del file */
	ret = fread(buffer, 1, size, file);
	  if(ret < size) {
	  printf("\n Errore nella lettura del file \n");
	  return 1;
	}
	
	fclose(file);
	
	/* Invio della lunghezza del nome del file */
	ret = send(sk, &name_size, sizeof(name_size), 0); 
	if(ret != sizeof(name_size)){
	  printf("\n Errore nella trasmissione della lunghezza del nome file\n ");
	  return 1;
	}
    
	/* Invio del nome del file */
	ret = send(sk, file_name, name_size, 0); 
	if(ret < name_size){
	  printf("\n Errore di trasmissione nome file\n ");
	  return 1;
	}
	
	/* allocazione del contesto */
	ctx = malloc(sizeof(EVP_CIPHER_CTX));
	
	if (create_enc_context(ctx, &block_size)) {
	    printf("Errore creazione contesto\n");
	    return 1;
	}
	
	/* allocazione del buffer per ciphertext */
	enc_buffer_size = size + block_size;
	
	enc_buffer = malloc(enc_buffer_size * sizeof (char));
	if (enc_buffer == NULL) {
	  printf("Errore allocazione buffer\n");
	  return 1;
	}
	
	/* cifratura del file */
	encrypt(ctx, buffer, size, enc_buffer, &ct_size);
	
	/* deallocazione contesto */
	EVP_CIPHER_CTX_cleanup(ctx);
	free(ctx);
	
	/* Invio della dimesione del file */
	ret = send(sk, &ct_size, sizeof(ct_size), 0);
	  if(ret != sizeof(ct_size)){
	  printf("\n Errore nella trasmissione della dimensione del file\n ");
	  return 1;
	}

	/* Invio del file */
	ret = send(sk, enc_buffer, ct_size, 0);
	if(ret < ct_size){
	  printf("\n Errore nella trasmissione del file \n");
	  return 1;
	}
	
	printf("\n Inviato il file %s di dimensione %d bytes, che cifrato diventa %d byte\n", file_name, size, ct_size);
	free(buffer);
	free(enc_buffer);
    
	return 0;
	
}

int main(int argc, char*argv[]) {   
  
    int ret;				/* Ritorno delle funzioni */
    int sk;				/* Socket di comunicazione col server */
    int cl_port; 			/* Numero di porta */

    struct sockaddr_in srv_addr;	/* Indirizzo del server */

    // Controllo degli argomenti in ingresso
    if (argc!=4) {
	printf ("Errore inserimento parametri. Usare: \n\t %s (IP) (porta) (nome_file)\n\n", argv[0]);
	return 1;
    }
    
    //Controllo numero di porta valido
    if ( atoi(argv[2]) <= 0 ||  atoi(argv[2]) > 65535 ) {
	printf ("Porta inserita non valida \n");
	return 1;
    }

    cl_port = atoi(argv[2]);
    printf ("server: Porta: %d \n",cl_port);
	
    memset(&srv_addr, 0, sizeof(srv_addr)); 
    srv_addr.sin_family = AF_INET; 
    srv_addr.sin_port = htons(cl_port); 
    ret = inet_pton(AF_INET, argv[1], &srv_addr.sin_addr);
    
    if(ret <= 0) {
        printf("\n Indirizzo del server errato\n");
        return 1;
    }
	  
    /* Creazione di un nuovo socket */
    sk = socket(AF_INET, SOCK_STREAM, 0);
    
    if(sk == -1) {
        printf("\n Non e' possibile creare un nuovo socket\n");
        return 1;
    }
    
    /* Creazione di una connessione TCP */
    ret = connect(sk, (SA *) &srv_addr, sizeof(srv_addr));
    
    if(ret == -1) {
        printf("\n Impossibile stabilire una connessione con il server\n");
        return 1;
    }
   
    printf("\n Connessione con il server %s (sulla porta %d).\n", argv[1], cl_port);

    /* invio del file al server */
    if(send_file(argv[3], sk))
	printf("Errore nell'invio del file al server!!\n");
   
    return 0;
    
}
