#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <openssl/evp.h>

#define SA struct sockaddr
#define BACKLOG_SIZE 64
#define SRV_ADDR "127.0.0.1"
    
void printbyte(char b) {
  
  char c;

  c = b;
  c = c >> 4;
  c = c & 15;
  printf("%X", c);
  c = b;
  c = c & 15;
  printf("%X:", c);
  
}

int retrieve_key(char* key, const int key_size) {
  
   int ret;
   int i;
   FILE* file;
   struct stat info;
   
   ret = stat("key", &info);
   if (ret != 0)
     return 1;
   
   file = fopen("key", "r");
   if(!file)
      return 1;
   
   ret = fread(key, 1, key_size, file);
   if(ret < key_size)
     return 1;

   fclose(file);
   return 0;
}

int create_dec_context(EVP_CIPHER_CTX *ctx, int* block_size) {
   
   int key_size; /* dimensione chiave di cifratura */
   char* key; /* chiave di cifratura */
    
   /* inizializzazione del contesto */
   EVP_CIPHER_CTX_init(ctx);
  
   /* setup del contesto per la cifratura */
   EVP_DecryptInit(ctx, EVP_des_ecb(), NULL, NULL);
  
   key_size = EVP_CIPHER_CTX_key_length(ctx);
   key = malloc(key_size);
   
   if (retrieve_key(key, key_size)) {
      printf("Errore recupero chiave\n");
      return 1;
   }
  
   /* impostazione della chiave */
   EVP_CIPHER_CTX_set_key_length(ctx, key_size);
   EVP_DecryptInit(ctx, NULL, (unsigned char*) key, NULL);
  
   /* recupero dimensione del blocco */
   *block_size = EVP_CIPHER_CTX_block_size(ctx);
  
   free(key);
   return 0;
  
}

int decrypt(EVP_CIPHER_CTX *ctx, char* buffer, char* dec_buffer, int dec_buffer_size, int* ct_size) {
  
  int nd; /* numero di byte decrittati ad ogni passo*/
  int ndtot; /* numero totale di byte decrittati */
  int i; /* indice */
  int ct_ptr, msg_ptr; /* puntatori alla prima posizione libera del buffer */
  int n;  /* numero di byte decifrare ad ogni passo */
    
  n = 10;
  nd = 0;
  ndtot = 0;
  ct_ptr = 0;
  msg_ptr =0;
      
  /* Decifratura a blocchi di n byte */
  /*for (i = 0; i < dec_buffer_size / n; i++) {
    EVP_DecryptUpdate(ctx, &buffer[ct_ptr], &nd, &dec_buffer[msg_ptr], n);
    ct_ptr += nd;
    msg_ptr += n;
    ndtot += nd;
  }
   
  if ( dec_buffer_size % n ) {
    EVP_DecryptUpdate(ctx, &buffer[ct_ptr], &nd, &dec_buffer[msg_ptr], dec_buffer_size % n);
    ct_ptr += nd;
    msg_ptr += dec_buffer_size % n;
    ndtot += nd;
  }*/
 
  /* Decifratura in un solo passaggio */
  EVP_DecryptUpdate(ctx, buffer, &nd, dec_buffer, dec_buffer_size);
  ct_ptr += nd;
  ndtot += nd;
  
  EVP_DecryptFinal(ctx, &buffer[ct_ptr], &nd);
  ndtot += nd;
  
  *ct_size = ndtot;
  
  return 0;
  
}
    
   /* Funzione per ricevere il file inviato dal client
   @ input sk = socket attraverso la quale si riceve il file
   @ return 0 in caso di successo, 1 altrimenti */
int receive_file(int sk) {

    int ret;
    int name_size;		// dimensione del nome del file ricevuto
    char* filename;		// nome del file ricevuto
    int size;			// dimensione del buffer per il testo in chiaro
    char* buffer;		// buffer per il testo in chiaro
    unsigned char* dec_buffer;	// buffer per il testo cifrato ricevuto
    int dec_buffer_size;	// dimensione del buffer per il testo cifrato ricevuto
    int block_size;		// dimensione del blocco
    int ct_size; 		// dimensione del testo in chiaro
	
    FILE* file;			// puntatore al file su cui salvare quanto ricevuto dal server
    EVP_CIPHER_CTX *ctx; 	// contesto

    /* Ricezione della lunghezza del nome del file */
    ret = recv(sk, &name_size, sizeof(name_size), MSG_WAITALL);
    if (ret != sizeof(name_size)) {
	  printf("%d \n Errore nella ricezione della dimensione del nome del file\n",ret);
	  return 1;
    }

    /* Allocazione della memoria */
    filename = malloc(sizeof(char) * (name_size + 1));
    if(filename == NULL) {
      printf("\n Errore nell'allocazione di memoria\n");
      return 1;
    }

    /* Ricezione del nome del file */
    ret = recv(sk, filename, name_size, MSG_WAITALL);
    
    if(ret != name_size){
      printf(" \n Errore nel ricevere il nome del file da trasferire\n");
      return 1;
    }
    
    filename[name_size] ='\0'; /* Terminazione stringa */
 
    /* Ricezione della dimensione del file */
    ret = recv(sk, &dec_buffer_size, sizeof(dec_buffer_size), MSG_WAITALL);
    if(ret != sizeof(dec_buffer_size)) {
      printf("\n Errore nel ricevere la dimensione del file\n");
      return 1;
    }
 
    /* Allocazione della memoria */
    dec_buffer = malloc(dec_buffer_size * sizeof(char));
    if(dec_buffer == NULL){
      printf("\n Errore nell'allocazione di memoria\n");
      return 1;
    }

    /* Ricezione del file */
    ret = recv(sk, dec_buffer, dec_buffer_size, MSG_WAITALL);
    if(ret != dec_buffer_size) {
      printf("\n Errore nel ricevere il file\n");
      return 1;
    }
       
    /* allocazione del contesto */
    ctx = malloc(sizeof(EVP_CIPHER_CTX));
       
    if (create_dec_context(ctx, &block_size)) {
	printf("Errore creazione contesto\n");
	return 1;
    }
	
    /* allocazione del buffer per il testo cifrato */
    size = dec_buffer_size + block_size;
   
    buffer = malloc(size * sizeof (char));
    if (buffer == NULL) {
	printf("Errore allocazione buffer\n");
	return 1;
    }

    /* decifratura */
    decrypt(ctx, buffer, dec_buffer, dec_buffer_size, &ct_size);
   
    /* deallocazione contesto */
    EVP_CIPHER_CTX_cleanup(ctx);
    free(ctx);
	
    /* Apertura del file in scrittura */
    file = fopen(filename,"w");
      if(file == NULL) {
	printf("\n File non trovato\n");
	return 1;
    }
    
    /* Chiamata alla fwrite() e riempimento del file */
    ret = fwrite(buffer, 1, ct_size, file);
    if(ret < ct_size) {
	printf("\n Errore nella scrittura del file \n");
	return 1;
    }    
    
    printf("\n Ricevuto il file %s di dimensione %d byte, che decifrato diventa %d byte\n", filename, dec_buffer_size, ct_size);
    
    fclose(file);
    free(filename);
    free(buffer);
    free(dec_buffer);
	
    return 0;
   
}
   
int main(int argc, char*argv[]) {
    
    socklen_t len;					/* Lunghezza dell'indirizzo del client */
    int sk;						/* Socket passivo */
    int optval;						/* Opzioni del socket */
    struct sockaddr_in my_addr, cl_addr;		/* Indirizzi del server e del client */
    char cl_paddr[INET_ADDRSTRLEN];			/* inidirizzo IP del client */
    uint16_t cl_port;					/* Porta del client */
    int ret;
    int cl_sk;						/* Client socket */
    int srv_port;					/* Porta del server */
    struct sockaddr_in srv_addr;			/* Indirizzo del server */
	
    printf("simple server, v. 0.1\n");   
	
    // Controllo degli argomenti in ingresso
    if (argc!=2) {
	printf ("Errore inserimento parametri usare: \n\t %s (porta) \n\n", argv[0]);
	return 1;
    }

    //Controllo numero di porta valido
    if (atoi(argv[1]) <= 0 || atoi(argv[1]) > 65535) {
	printf ("Porta inserita non valida \n");
	return 1;
    }
	
    srv_port = atoi(argv[1]);
    printf ("server: Porta: %d \n",srv_port);
	
    memset(&srv_addr, 0, sizeof(srv_addr)); 
    srv_addr.sin_family = AF_INET; 
    srv_addr.sin_port = htons(srv_port); 
    ret = inet_pton(AF_INET, SRV_ADDR, &srv_addr.sin_addr);
    
    if(ret <= 0) {
        printf("\n Indirizzo del server errato\n");
        return 1;
    }
 
    /* Creazione di un nuovo socket */
    sk = socket(AF_INET, SOCK_STREAM, 0);
    if(sk == -1){
        printf("\n Impossibile creare un nuovo socket\n");
        return 1;
    }
    
    optval = 1;
    ret = setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
    if(ret == -1) {
        printf("\n Errore nel settaggio di SO_REUSEADDR \n");
        return 1;
    }
    
    /* Collegamento del socket all'indirizzo IP e alla porta */
    memset(&my_addr, 0, sizeof(my_addr)); 
    my_addr.sin_family = AF_INET;
    my_addr.sin_addr.s_addr = htonl(INADDR_ANY); 
    my_addr.sin_port = htons(srv_port);
    
    ret = bind(sk, (SA *) &my_addr, sizeof(my_addr));
    if(ret == -1) {
        printf("\n Impossibile collegare il socket\n");
        return 1;
    }
    
    /* Creazione della coda di backlog */
    ret = listen(sk, BACKLOG_SIZE);
    if(ret == -1) {
        printf("\n Impossibile creare la coda di backlog lunga %d\n", BACKLOG_SIZE);
        return 1;
    }
    	
    printf("In attesa di connessioni....\n");
    
    while(1) {
        
	/* Accetta una richiesta arrivata su sk e la fa gestire a cl_sk */
        len = sizeof(cl_addr);
        cl_sk = accept(sk, (SA *) &cl_addr, &len);
        
	if(cl_sk == -1) {
            printf("\n Errore durante la connessione \n");
            return 1;
        }
        
        inet_ntop(AF_INET, &cl_addr.sin_addr, cl_paddr, sizeof(cl_paddr));
        cl_port = ntohs(cl_addr.sin_port);
        printf("\n Connessione al client %s, sulla porta %d avvenuta \n", SRV_ADDR, srv_port);

	if(receive_file(cl_sk)) {
	  printf("Errore nella ricezione del file \n");
	  return 1;
	}
		
	printf ("\n Termine della connessione\n");
	close(cl_sk);
	return 0;
    }
	
   close(sk);
   return 0;
   
}
